/*
 * DbObject.java
 * 2002-07-08  Knabe  oid state evolves from isCreated() to isPersistent to isDeleted
 * 2002-04-30  Knabe  dbUpdate() checks new db__updateVersion counter.
 * 2002-04-11  Knabe  All modifying operation names begin with prefix db
 * 2001-05-28  Knabe  Preventing duplicate insert.
 *   All mutator operations only as protected, against usage from UI tier.
 * 2001-05-03  Knabe  OID type changed from int to Integer
 * 2001-04-22  Knabe  All object related DB operations here as protected
 * Created on 28. Februar 2001, 15:31
 */

package bank3tier.db;

/**
 *
 * @author  Christoph Knabe
 * @version
 */
/**Base class for all persistent objects. This class is purely generic.
  All application specific code (especially table creation) is collected
  in a subclass of {@link DbApplication}.
  Each object is identified by an OID. The OID (Object IDentification) is a
  number unique for each object in our application.
  This class supports the OR-Mapping according to Scott Ambler. But his
  retrieve-Operation is not implemented, as it is only necessary for
  efficiency optimization. And his save-Operation is splitted to the more
  common insert() and update().
  The state of a DbObject evolves from isCreated() by dbInsert() to
  isPersistent() and by dbDelete() to isDeleted().
*/
public abstract class DbObject {

    public static final String oidSuffix = "_oid";

    /**Returns the OID of this object.*/
    public final Integer getOid(){return oid;}


    /**Returns true, if this object has been newly created without being made persistent*/
    public final boolean isCreated(){return oid==_createdOid;}

    /**Returns true, if this object has been made persistent*/
    public final boolean isPersistent(){return oid!=null && oid.intValue() > 0;}

    /**Returns true, if this object has been deleted from the persistence mechanism*/
    public final boolean isDeleted(){return oid==_deletedOid;}

    /**Opens the database and sets the necessary modes*/
    public static void dbOpenDB(final DbApplication io_dbApplication)
    {
        try{
            io_dbApplication.openDB();
            io_dbApplication.getSession().connection.setAutoCommit(false);
            _dbApplication = io_dbApplication;
        }catch(Exception ex){
            throw new multex.Failure("Failure opening database", ex);
        }
    }

    /**Closes the database and releases its resources*/
    public static void dbCloseDB()
    {
		_dbApplication.getSession().close();
        _dbApplication = null;
    }

    /**Creates and returns a new Table object for managing the access to the table
      for storing objects of class i_class.
    */
    protected static jora.Table dbNewTable(final Class i_class){
        final String className = i_class.getName();
        final int localClassNameIndex = className.lastIndexOf('.') + 1;
        final String tableName
        = className.substring(localClassNameIndex+2).toLowerCase();//without "Db"
        return new jora.Table(
            className, tableName, _dbApplication.getSession(), "oid"
        );
    }

    /**Creates the application specific database tables*/
    public static void dbCreateTables() throws multex.Failure {
        try{
            _dbApplication.createTables(oidSuffix);
            dbEndTransaction(true);
        }catch(Exception ex){
            throw new multex.Failure("Failure creating the application "
                + "specific database tables", ex
            );
        }
    }

    /**Deletes the application specific database tables*/
    protected static void dbDeleteTables() throws multex.Failure {
        try{
            final jora.Session session = _dbApplication.getSession();
            session.closePreparedStatements();
            _dbApplication.deleteTables();
            dbEndTransaction(true);
            session.close();
        }catch(Exception ex){
            throw new multex.Failure("Failure deleting the application "
                + "specific database tables", ex
            );
        }

    }

    /** Insert this object into its table in the database.
      Before inserting the object gets a new OID.
      @throws DuplicateInsertExc The actual object was already in the database
    */
   public void dbInsert() throws DuplicateInsertExc
    {
      if(isPersistent()){throw new DuplicateInsertExc(dbTable().name(), getOid());}
      oid = _getNewOid();   dbTable().insert(this);
    }

    /** Retrieves this object from its table in the database.
      Checks that the object has not been modified in the database
      after the last select() from the database.
    . You should in a transaction firstly call
      this method for each object participating in the transaction.
   public void dbRetrieve()
    {
        final jora.Table table = dbTable();
        try{
            final DbObject dbObject = dbSelectByOid(table, getOid());
            //table.copy(dbObject, this); //Now do no longer overwrite
            //contents of the actual object.
        } catch(Exception ex){
            throw new multex.Failure(
                "Failure retrieving object with OID {0} from table {1}",
                ex, getOid(), table.name()
            );
        }
    }
    */

    /** Update the representation in the database from this object.
     * @return number of objects actually updated
    */
    public final int dbUpdate() throws ConcurrentUpdateExc
    {
		final jora.Table table = dbTable();
        try{
            final DbObject dbObject = dbSelectByOid(table, getOid());
            //old approach: table.copy(dbObject, this);
            //new approach:
            if(dbObject.db__updateVersion!=this.db__updateVersion){ //meanwhile updated in the database
                throw new ConcurrentUpdateExc(table.name(), getOid());
            }

            db__updateVersion++;
            return table.update(this);
        } catch(ConcurrentUpdateExc ex){throw ex;
        } catch(Exception ex){
            throw new multex.Failure(
                "Failure updating object with OID {0} in table {1}",
                ex, getOid(), table.name()
            );
        }
    }

    /** Delete the representation of this object in the database.
     * Additonally mark this object as not valid by giving it the OID _noOid.
     * @return number of objects actually deleted
     */
    public final int dbDelete()
    { final int rowCount = dbTable().delete(this); oid = _deletedOid; return rowCount;}

    /** Returns the object with the OID i_oid from the database table i_table.
     *  The returned object will be dynamically an instance of the type given
     *  as first parameter to the constructor of i_table.
     *  Note: The selected object should
     *  have no components of InputStream, Blob or Clob type, because
     *  their data will be not available after closing the Cursor used to get
     *  it.
     *
     *  @throws OidNotFoundExc If in i_table no object with OID i_oid was found.
     */
    protected static DbObject dbSelectByOid(final jora.Table i_table,final Integer i_oid) throws OidNotFoundExc
    {
        final jora.Cursor c = i_table.selectAll("where oid="+i_oid);
        final Object result = c.next();
        if(result==null){
            throw new OidNotFoundExc(i_oid, i_table);
        }
        //???Wegen Verlust der BLOBS usw. so sinnvoll?
        c.close();
        return (DbObject)result;
    }

    /** Returns the one unique object by a "query by example"
     *  from the database.
     *  The returned object will be dynamically an instance of the same type
     *  as the actual example object.
     *  Note: The selected object should
     *  have no components of InputStream, Blob or Clob type, because
     *  their data will be not available after closing the Cursor used to get
     *  it.
     *
     *  @throws NotFoundExc If no object matching this example object was found.
     *  @throws NotUniqueExc If more than one object matches this example object.
     */
    protected DbObject dbSelectUniqueByExample() throws ExampleNotFoundExc, ExampleNotUniqueExc
    {
        final jora.Table table = dbTable();
        final jora.Cursor cursor = table.queryByExample(this);
        final Object[] allObjects = cursor.toArray();
        final String query = cursor.getQuery(); //Query only known after Cursor.next()
        if(allObjects.length>1){throw new ExampleNotUniqueExc(table, query);
        }else if(allObjects.length<1){
            throw new ExampleNotFoundExc(table, query); //???Parameter
        }
        return (DbObject)allObjects[0];
    }

    /**Returns the DbObject of the the class i_partnerClass, which is linked
      to from the actual object. The name of the class i_partnerClass is used
      to determine a) the Table, in which to serach, and b) the value of the
      OID looked for is taken from the own attribute, whose name is formed
      as a concatenation of the partner class name, and the suffix oidSuffix.
    */
    public DbObject dbGetlink(final Class i_partnerClass) throws LinkFailure
    {
        final Class ownClass = getClass();
        String partnerTableName = "";
        String partnerOidFieldName = "";
        Integer partnerOid = null;
        try {
            //Get the Table of i_partnerClass:
            final java.lang.reflect.Field partnerTableField
            = i_partnerClass.getDeclaredField("table");
            partnerTableField.setAccessible(true);
            final Object partnerTableObject = partnerTableField.get(null);
            final jora.Table partnerTable = (jora.Table)partnerTableObject;
            //Determine the OID of the linked partner object:
            partnerTableName = partnerTable.name();
            partnerOidFieldName = partnerTableName.toLowerCase() + oidSuffix;
            final java.lang.reflect.Field partnerOidField
            = ownClass.getDeclaredField(partnerOidFieldName);
            partnerOidField.setAccessible(true);
            final Object partnerOidObject = partnerOidField.get(this);
            partnerOid = (Integer)partnerOidObject;
            //Get the object of the partner table with the partner OID:
            return dbSelectByOid(partnerTable, partnerOid);
        } catch(Exception ex){
            throw new LinkFailure(ex, ownClass.getName(), partnerOidFieldName, partnerTableName, partnerOid);
        }
    }

    /**Returns true, if in the actual table there is a pair with the oid-s
      of the DbObject-s i_left and i_right as foreign keys
    */
    public boolean dbAreLinked(final DbObject i_left,final DbObject i_right) {
        final Object[] links = dbTable().select(
            "WHERE " + i_left.dbTable().name() + oidSuffix + " = " + i_left.getOid()
            + " AND "+ i_right.dbTable().name() + oidSuffix + " = "+i_right.getOid()
        ).toArray();
        return links.length >= 1;
    }

    /** Mark the beginning of an transaction.
      Unconfirms all modifications made to the database after the last
      call to {@link #endTransaction(boolean)}.
     */
    public static void dbBeginTransaction()
    { _dbApplication.getSession().rollback(); }

    /** Confirm or unconfirm the modifications made to the database in the
      actual transaction (since the last commit) and close this transaction.
      @param i_ok true if the transaction succeeded, then the modifications
        will be confirmed, otherwise unconfirmed.
     */
    public static void dbEndTransaction(final boolean i_ok)
    {
        final jora.Session session = _dbApplication.getSession();
        if(i_ok){session.commit();}else{session.rollback();}
    }


    public class ConcurrentUpdateExc extends multex.Exc {
      public ConcurrentUpdateExc(final String i_tableName, final Integer i_oid){
        super("{0}-object with OID {1} has been modified by another user. "
          + " You should repeat the transaction", i_tableName, i_oid
        );
      }
    }

    public class DuplicateInsertExc extends multex.Exc {
      public DuplicateInsertExc(final String i_tableName, final Integer i_oid){
        super("Cannot insert {0}-object with OID {1}, "
          + "as it is already in the database", i_tableName, i_oid
        );
      }
    }

    public static class OidNotFoundExc extends multex.Exc {
        public OidNotFoundExc(final Integer i_oid,final jora.Table i_table){
            super("Object with OID {0} not found in table \"{1}\"",
                i_oid, i_table.name()
            );
        }
    }

    public static class ExampleNotFoundExc extends multex.Exc {
        public ExampleNotFoundExc(final jora.Table i_table, final String i_sql){
            super("Object not found by example in table \"{0}\", 1=sql",
                i_table.name(), i_sql
            );
        }
    }

    public static class ExampleNotUniqueExc extends multex.Exc {
        public ExampleNotUniqueExc(final jora.Table i_table, final String i_sql){
            super("Object selection by example not unique in table \"{0}\", 1=sql",
                i_table.name(), i_sql
            );
        }
    }

    public static class LinkFailure extends multex.Failure {
        public LinkFailure(final Exception i_cause, final String i_ownClassName, final String i_partnerOidFieldName, final String i_partnerTableName, final Integer i_partnerOid){
            super("Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3} ",
                i_cause, i_ownClassName, i_partnerOidFieldName, i_partnerTableName, i_partnerOid
            );
        }
    }

    /** Get the JORA table responsible for storing objects of the same class
      as this.
      It is necessary to redefine this method in each concrete class by <PRE>
      *   jora.Table table(){return table;}
      </PRE>
    */
    protected abstract jora.Table dbTable();

    /**Unique Object IDentification for each DbObject instance*/
    private Integer oid = _createdOid;

    /**Modification version counter for this object, shall be incremented
      by any insert oder update operation.
    */
    private int db__updateVersion = 0;

    private static DbApplication _dbApplication;

    private static int _lastOid = -1;
    private static final Integer _deletedOid = _getNewOid();
    private static final Integer _createdOid = null;

    /** Returns an object identification number unique for each object in our
      * application. In a non-demo application with real persistency the
      * _lastOid should be stored in the database, too, rather than in a
      * variable.
    */
    private static Integer _getNewOid(){return new Integer(++_lastOid);}


}